#include <math.h>
#include "llsmapping.h"
#include <stdio.h>
#include <iostream>
#include <fstream>
#include "const.h"
#include "paramSimul.h"
using namespace std;

double Mapping_RBIR_SNR(int mod_idx, double input, int mode) // mode 0: SNR -> RBIR, mode 1: RBIR -> SNR   input-dB
{
	int i, data_range;
	double in_SNR, out_SNR;
	double in_RBIR, out_RBIR;
	double SNR[95];
	for (int i = 0; i<95; i++)	SNR[i] = 0.5*i - 20.;

	data_range = 95;
	

	if (mode == 0) {///// SNR -> RBIR
		in_SNR = input;

		if (in_SNR<SNR[0]) // outerpolation
		{
			out_RBIR = ((SNR[1] - in_SNR)*RBIR[mod_idx][0] - (SNR[0] - in_SNR)*RBIR[mod_idx][1]) / (SNR[1] - SNR[0]);
			if (out_RBIR < 0) out_RBIR = 0;
		}

		else if (in_SNR >= SNR[data_range - 1])
			out_RBIR = 1.;

		else { // interpolation
			for (i = 0; i<data_range - 1; i++) {
				if (SNR[i] <= in_SNR && SNR[i + 1]>in_SNR) {
					out_RBIR = ((RBIR[mod_idx][i] - RBIR[mod_idx][i + 1])*in_SNR + SNR[i] * RBIR[mod_idx][i + 1] - SNR[i + 1] * RBIR[mod_idx][i]) / (SNR[i] - SNR[i + 1]);
				}
			}
		}
		return out_RBIR;
	}
	else {///// RBIR -> SNR
		in_RBIR = input;

		if (in_RBIR > 1.00001)
		{
			cout << "check in_RBIR" << endl;
			getchar();
		}

		if (in_RBIR<RBIR[mod_idx][0]) // outerpolation
			out_SNR = ((RBIR[mod_idx][1] - in_RBIR)*SNR[0] - (RBIR[mod_idx][0] - in_RBIR)*SNR[1]) / (RBIR[mod_idx][1] - RBIR[mod_idx][0]);

		else if (in_RBIR >= RBIR[mod_idx][data_range - 1]) { // the smallest SNR whose RBIR value is one (specific value according to RBIR-SNR mapping table)
			if (mod_idx == 0) out_SNR = SNR[65];
			else if (mod_idx == 1) out_SNR = SNR[79];
			else if (mod_idx == 2) out_SNR = SNR[91];
		}

		else { // interpolation
			for (i = 0; i<data_range - 1; i++) {
				if (RBIR[mod_idx][i] <= in_RBIR && RBIR[mod_idx][i + 1]>in_RBIR) {
					out_SNR = ((SNR[i] - SNR[i + 1])*in_RBIR + RBIR[mod_idx][i] * SNR[i + 1] - RBIR[mod_idx][i + 1] * SNR[i]) / (RBIR[mod_idx][i] - RBIR[mod_idx][i + 1]);
				}
			}
		}
		return  out_SNR;
	}
}

double Determine_TBS(int num_layer, int cqi, int n_PRB)
{
	double Spectral_efficiency = Get_SpectralEfficiency(cqi);
	double R = Get_code_rate(cqi);

	double N_RE_prime = 12 * 14 * (1 - overhead);
	double N_RE = N_RE_prime * n_PRB;
	double N_info = N_RE * Spectral_efficiency*num_layer;
	double N_info_prime;
	double n;
	double TBS;
	double C;

	if (N_info <= 3824.)
	{
		n = MAX(3.0, floor(log2(N_info)) - 6.0);
		N_info_prime = MAX(24., pow(2.0, n)*floor(N_info / pow(2., n)));

		for (int i = 0; i < 92; i++)
		{
			if (TBS_table[91 - i] < N_info_prime)
			{
				TBS = TBS_table[92 - i];
				break;
			}
		}
	}
	else
	{
		n = floor(log2(N_info-24.))-5. ;
		N_info_prime = MAX(3840., pow(2.0, n) * floor((N_info - 24.) / pow(2.0, n) +0.5));

		if (R <= 1. / 4.)
		{
			C = ceil((N_info_prime + 24.) / 3816.);
			TBS = 8 * C*ceil((N_info_prime + 24.) / (8.*C));
			
		}
		else
		{
			if (N_info_prime > 8424)
			{
				C = ceil((N_info_prime + 24.) / 8424);
				TBS = 8 * C*ceil((N_info_prime + 24.) / (8.*C)) - 24;
			}
			else
			{
				TBS = 8 * ceil((N_info_prime + 24.) / 8.) - 24;
			}
		}

	}
	return TBS;
}

double Get_SpectralEfficiency(int cqi)
{
	double Spectral_efficiency;
	if (cqi == -1) { Spectral_efficiency = 0.0; }
	else if (cqi == 0) { Spectral_efficiency = 0.1523; }
	else if (cqi == 1) { Spectral_efficiency = 0.2344; }
	else if (cqi == 2) { Spectral_efficiency = 0.3770; }
	else if (cqi == 3) { Spectral_efficiency = 0.6016; }
	else if (cqi == 4) { Spectral_efficiency = 0.8770; }
	else if (cqi == 5) { Spectral_efficiency = 1.1758; }
	else if (cqi == 6) { Spectral_efficiency = 1.4766; }
	else if (cqi == 7) { Spectral_efficiency = 1.9141; }
	else if (cqi == 8) { Spectral_efficiency = 2.4063; }
	else if (cqi == 9) { Spectral_efficiency = 2.7305; }
	else if (cqi == 10) { Spectral_efficiency = 3.3223; }
	else if (cqi == 11) { Spectral_efficiency = 3.9023; }
	else if (cqi == 12) { Spectral_efficiency = 4.5234; }
	else if (cqi == 13) { Spectral_efficiency = 5.1152; }
	else if (cqi == 14) { Spectral_efficiency = 5.5547; }
	else
	{
		cout << "check Spectral_efficiency" << endl;
		getchar();
	}

	return Spectral_efficiency;
}

int Map_CQI(double SINR)
{
	int MCS;
	if (SINR < -7.8474) { MCS = -1; }
	else if (SINR >= -7.8474&&SINR < -6.2369) { MCS = 0; }
	else if (SINR >= -6.2369&&SINR < -4.3591) { MCS = 1; }
	else if (SINR >= -4.3591&&SINR < -1.9319) { MCS = 2; }
	else if (SINR >= -1.9319&&SINR < 0.1509) { MCS = 3; }
	else if (SINR >= 0.1509&&SINR < 1.9976) { MCS = 4; }
	else if (SINR >= 1.9976&&SINR < 4.7278) { MCS = 5; }
	else if (SINR >= 4.7278&&SINR < 6.2231) { MCS = 6; }
	else if (SINR >= 6.2231&&SINR < 8.0591) { MCS = 7; }
	else if (SINR >= 8.0591&&SINR < 9.8585) { MCS = 8; }
	else if (SINR >= 9.8585&&SINR < 11.8432) { MCS = 9; }
	else if (SINR >= 11.8432&&SINR < 13.4893) { MCS = 10; }
	else if (SINR >= 13.4893&&SINR < 15.3598) { MCS = 11; }
	else if (SINR >= 15.3598&&SINR < 17.4435) { MCS = 12; }
	else if (SINR >= 17.4435&&SINR < 19.2155) { MCS = 13; }
	else if (SINR >= 19.2155) { MCS = 14; }
	return MCS;
}

double Get_code_rate(int cqi)
{
	double coderate;
	if (cqi == -1) { coderate = 0.0; }
	else if (cqi == 0) { coderate = 0.1523/2; }
	else if (cqi == 1) { coderate = 0.2344/2; }
	else if (cqi == 2) { coderate = 0.3770/2; }
	else if (cqi == 3) { coderate = 0.6016/2; }
	else if (cqi == 4) { coderate = 0.8770/2; }
	else if (cqi == 5) { coderate = 1.1758/2; }
	else if (cqi == 6) { coderate = 1.4766/4; }
	else if (cqi == 7) { coderate = 1.9141/4; }
	else if (cqi == 8) { coderate = 2.4063/4; }
	else if (cqi == 9) { coderate = 2.7305/6; }
	else if (cqi == 10) { coderate = 3.3223/6; }
	else if (cqi == 11) { coderate = 3.9023/6; }
	else if (cqi == 12) { coderate = 4.5234/6; }
	else if (cqi == 13) { coderate = 5.1152/6; }
	else if (cqi == 14) { coderate = 5.5547/6; }
	else
	{
		cout << "check Spectral_efficiency" << endl;
		getchar();
	}

	return coderate;
}

int Get_modulation_Scheme(int cqi)
{
	int mod_idx;
	if (cqi < 6)
	{
		mod_idx = 0; //QPSK
	}
	else if (cqi < 9)
	{
		mod_idx = 1; //16QAM
	}
	else
	{
		mod_idx = 2;//64QAM
	}
	return mod_idx;
}
