

#include <stdio.h>
#include <iostream>
#include <fstream>
#include "sector.h"
#include "paramSimul.h"
#include "evaluationConfigurerations.h"
#include "ue.h"
#include "const.h"
#include "struct.h"
#include "objectBSSectorUe.h"
#include "channel.h"

/*===================================================================
FUNCTION: Sector::Sector()

AUTHOR: SJ

DESCRIPTION: Sector Object Constructor

NOTES: Currently, nothing is done when a sector is instantiated

===================================================================*/

Sector::Sector()
{
	pschedulewrite = new SCHEDULE_DECISION[num_rb];
	cqis = NULL;

}

/*===================================================================
FUNCTION: Sector::~Sector()

AUTHOR: SJ

DESCRIPTION: Sector Object Destructor

NOTES: Currently, nothing is done

===================================================================*/
Sector::~Sector()
{
	delete pschedulewrite;
}

/*===================================================================
FUNCTION: Sector::Configure( int )

AUTHOR: SJ

DESCRIPTION: Sector object gets to know the index of itself

NOTES: More functionality may be added

===================================================================*/
void Sector::initialization(int sector_idx)
{
	ue_in_control.clear();
	
	_sector_idx = sector_idx;

	if (num_TRxP_per_site == 3)
	{
		alpha = (sector_idx % 3 * 120 + 30)*PI / 180;
		
	}
	else
	{
		alpha = 0;
	}
	
	beta = (bs_antenna.mechanic_tilt - 90) * PI / 180;
	gamma = 0;
}

/*===================================================================
FUNCTION: Sector::RR_scheduling()

AUTHOR: GJ

DESCRIPTION:

OUTPUT: pschedulewrite[rb_idx].ue_selected

NOTES: More functionality may be added

===================================================================*/

void Sector::RR_scheduling()
{

	//cout << "ramdom_selection_UE" << endl;
	ramdom_selection_UE();

	if (t <csi_feedback_interval)
	{
		ramdom_selection_RI_PMI();
	}
	else 
	{
		for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
		{	
			pschedulewrite[rb_idx].RI = ue[pschedulewrite[rb_idx].ue_selected].RI;
			//cout << "pschedulewrite[rb_idx].RI : " << pschedulewrite[rb_idx].RI << endl;
			pschedulewrite[rb_idx].PMI = ue[pschedulewrite[rb_idx].ue_selected].PMI[rb_idx][t % csi_history_length];	
			//cout << "pschedulewrite[rb_idx].PMI" <<pschedulewrite[rb_idx].PMI << endl;
		}
	}
	   
	
	//determine_MCS();
	UE_Power_control();
}

void Sector::ramdom_selection_UE()
{

	SCHEDULE_DECISION *decision;
	decision = new SCHEDULE_DECISION[num_rb];


	for (int num_rb = 0; num_rb < num_rb; num_rb++)
	{
		decision[num_rb].ue_selected = -1;
	}

	
	for (int num_rb = 0; num_rb < num_rb; num_rb++)
	{
		decision[num_rb].ue_selected = ue_in_control[(int)(Rand.u()*ue_in_control.size())];
	}

	

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		if (t >5)
		{
			/*cout << " pschedulewrite["<<rb_idx<<"].ue_selected : "<<pschedulewrite[rb_idx].ue_selected << endl;
			getchar();*/
			if (ue[pschedulewrite[rb_idx].ue_selected].ack == ACK)
			{
				pschedulewrite[rb_idx].ue_selected = decision[num_rb].ue_selected;
			}
		}
		else
		{
			pschedulewrite[rb_idx].ue_selected = decision[num_rb].ue_selected;
		}
	
	}

	delete decision;
}
void Sector::ramdom_selection_RI_PMI()
{
	bool count = false;

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		pschedulewrite[rb_idx].RI = num_ue_antenna_port-1;
		pschedulewrite[rb_idx].CQI_selected = int(Rand.u() * 16)-1;
	}
	
	
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		if (pschedulewrite[rb_idx].RI == 0)
		{
			pschedulewrite[rb_idx].PMI = int(Rand.u() * 28);
		}
		else if (pschedulewrite[rb_idx].RI == 1)
		{
			pschedulewrite[rb_idx].PMI = int(Rand.u() * 22);
		}
		else if (pschedulewrite[rb_idx].RI == 2)
		{
			pschedulewrite[rb_idx].PMI = int(Rand.u() * 7);
		}
		else if (pschedulewrite[rb_idx].RI == 3)
		{
			pschedulewrite[rb_idx].PMI = int(Rand.u() * 5);
		}
		else
		{
			cout << "check pschedulewrite[rb_idx].RI" << endl;
			cout << "pschedulewrite["<<rb_idx<<"].RI : "<< pschedulewrite[rb_idx].RI << endl;
			getchar();
		}
	}
}
void Sector::determine_CQI()
{
	
}

void Sector::UE_Power_control()
{
	int ue_selected;
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		ue_selected = pschedulewrite[rb_idx].ue_selected;
		ue[ue_selected].Find_Allocated_Rbs();
		ue[ue_selected].power_per_RB = ue[ue_selected].Update_Power_Control(ue[ue_selected].num_of_rbs);
	}
}


void Sector::PF_scheduling()
{
	SCHEDULE_DECISION *decision;
	decision = new SCHEDULE_DECISION[num_rb];

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		decision[rb_idx].ue_selected = -1;
	}
		
	int max_metric_ue = 0;
	double T;
	double R;
	double max_metric=0;
	double metric = 0;
	int max_rs = 7;

	vector <int>scheduling_candidate;
	
	if (double(num_rb)/double(ue_in_control.size()) >max_rs)
	{
		max_rs = ceil(num_rb / int(ue_in_control.size()));
	}

	if (ue_in_control.size() == 0)
	{
		cout << "check ue_in_control.size()  " << endl;
		cout << "ue_in_control.size() : " << ue_in_control.size() << endl;
		getchar();
	}

	
	for (int control_ue_idx = 0; control_ue_idx < ue_in_control.size(); control_ue_idx++)
	{
		
		if (ue[ue_in_control[control_ue_idx]].ack == ACK)
		{
			ue[ue_in_control[control_ue_idx]].rb_indices.clear();
			scheduling_candidate.push_back(ue_in_control[control_ue_idx]);
		}
	}

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		if (ue[pschedulewrite[rb_idx].ue_selected].ack == ACK)
		{
			//cout << "scheduling_candidate.size() : " << scheduling_candidate.size() << endl;
			for (int control_ue_idx = 0; control_ue_idx < scheduling_candidate.size(); control_ue_idx++)
			{				
				R = ue[scheduling_candidate[control_ue_idx]].R_for_pf[rb_idx];
				T = ue[scheduling_candidate[control_ue_idx]].T_for_pf[rb_idx];
				if (T == 0)
				{
					T = 0.00000000001;
				}

				metric = R / T;

				/*if (_sector_idx == 0)
				{
					cout << "ue["<<ue_in_control[control_ue_idx]<<"].R_for_pf : "<< ue[ue_in_control[control_ue_idx]].R_for_pf  << endl;
					cout << "ue[" << ue_in_control[control_ue_idx] << "].T_for_pf : " << ue[ue_in_control[control_ue_idx]].T_for_pf << endl;
					cout << "metric : " << metric << endl;
				}*/

				if (control_ue_idx == 0)
				{
					max_metric_ue = control_ue_idx;
					max_metric = metric;
				}
				else if (max_metric<metric)
				{
					max_metric_ue = control_ue_idx;
					max_metric = metric;
				}
			}

			ue[scheduling_candidate[max_metric_ue]].rb_indices.push_back(rb_idx);
			
			decision[rb_idx].ue_selected = scheduling_candidate[max_metric_ue];
			
			for (int control_ue_idx = 0; control_ue_idx < ue_in_control.size(); control_ue_idx++)
			{
				if (control_ue_idx == max_metric_ue)
				{
					ue[ue_in_control[control_ue_idx]].T_for_pf[rb_idx] = (1. - 1. / (t - 4.))*ue[ue_in_control[control_ue_idx]].T_for_pf[rb_idx] + 1. / (t - 4.)* ue[ue_in_control[control_ue_idx]].R_for_pf[rb_idx];
				}
				else
				{
					ue[ue_in_control[control_ue_idx]].T_for_pf[rb_idx] = (1.-1./ (t - 4.))*ue[ue_in_control[control_ue_idx]].T_for_pf[rb_idx];
				}
				//if(_sector_idx==0)  cout<<" ue["<<ue_in_control[control_ue_idx]<<"].T_for_pf : " <<ue[ue_in_control[control_ue_idx]].T_for_pf<<endl;
			}
			
			if (ue[scheduling_candidate[max_metric_ue]].rb_indices.size() == max_rs)
			{
				scheduling_candidate.erase(scheduling_candidate.begin() + max_metric_ue);
			}


			//if (_sector_idx == 0) cout << "max_metric_ue : "<< ue_in_control[max_metric_ue] << endl;


			
		}
		else
		{
			decision[rb_idx].ue_selected = pschedulewrite[rb_idx].ue_selected;
		}
	}



	/*cout << endl;
	cout << "sec_idx : " << _sector_idx << endl;
*/
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		pschedulewrite[rb_idx].ue_selected = decision[rb_idx].ue_selected;
		//cout << "pschedulewrite["<<rb_idx<<"].ue_selected" << pschedulewrite[rb_idx].ue_selected << endl;
	}


	/*
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		if (_sector_idx == 0) cout <<"pschedulewrite["<<rb_idx<<"].ue_selected "<<pschedulewrite[rb_idx].ue_selected << endl;
	}

	getchar();
	*/

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		//cout << "pschedulewrite[rb_idx].CQI_selected : " << pschedulewrite[rb_idx].CQI_selected << endl;
		pschedulewrite[rb_idx].RI = ue[pschedulewrite[rb_idx].ue_selected].RI;
		//cout << "pschedulewrite[rb_idx].RI : " << pschedulewrite[rb_idx].RI << endl;
		pschedulewrite[rb_idx].PMI = ue[pschedulewrite[rb_idx].ue_selected].PMI[rb_idx][t % csi_history_length];
		//cout << "pschedulewrite[rb_idx].PMI" <<pschedulewrite[rb_idx].PMI << endl;
	}

	delete decision;
	
}



