#include "ue.h"
#include <stdio.h>
#include <iostream>
#include <fstream>
#include "const.h"
#include "struct.h"
#include "paramSimul.h"
#include "paramFastFading.h"
#include "evaluationConfigurerations.h"
#include "objectBSSectorUe.h"
#include "channel.h"
#include "llsmapping.h"
using namespace std;

/*===================================================================
FUNCTION: UE::UE()

AUTHOR: SJ

DESCRIPTION: UE Object Constructor

NOTES: Currently, nothing is done when a UE is instantiated

===================================================================*/
UE::UE()
{
	adj_bs = new int[num_compute_coef];
	
	PMI = new  int*[num_rb];
	CQI = new  int*[num_rb];
	
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		PMI[rb_idx] = new int[csi_history_length];
		CQI[rb_idx] = new int[csi_history_length];
	}

}

/*===================================================================
FUNCTION: UE::~UE()


AUTHOR: SJ

DESCRIPTION: UE Object Destructor

NOTES: Currently, nothing is done

===================================================================*/
UE::~UE()
{	

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		delete[] PMI[rb_idx];
		delete[] CQI[rb_idx];
	}

	delete[] adj_bs;
}


/*===================================================================
FUNCTION: UE::Configure( int ue_idx )

AUTHOR: JH

DESCRIPTION: UE object gets to know the index of itself

NOTES:

===================================================================*/
void UE::Configure(int ue_idx)
{

	Pathloss_car = 0;
	Pathloss_tw = 0;
	Set_UE_speed_indoor_outdoor_state();
	Set_h_UT();
	_ue_idx = ue_idx;
	uplinkThroghput = 0;
	HARQeSINRs.clear();
	num_retransmission = 0;
	ack = ACK;

	alpha = Rand.u()*2.0*PI - PI;
	beta = 0;
	gamma = 0;
	
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		T_for_pf[rb_idx] = 0;
		R_for_pf[rb_idx] = 0;
	}
	

	num_transmission=0;
	num_error[0]=0;
	num_error[1] = 0;
	num_error[2] = 0;
	num_error[3] = 0;
}


void UE::Allocate_memory()
{


}


void UE::Set_UE_speed_indoor_outdoor_state()
{
	if (test_environment <3) //Indoor
	{
		speed = indoor_ue_speed * 1000. / 3600.;
	}
	else if (test_environment<9) // UMa, UMi
	{
		if (Rand.u() < indoor_ue_portion) // if (Rand.u() < indoor_ue_portion)  indoor outdoor 
		{
			ue_indoor_outdoor_state = INDOOR;
			
			speed = indoor_ue_speed * 1000. / 3600.;
		}
		else
		{
			ue_indoor_outdoor_state = IN_CAR;

			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}		
	}
	else //Rural
	{
		if (Rand.u() < indoor_ue_portion)
		{
			ue_indoor_outdoor_state = INDOOR;
			
			speed = indoor_ue_speed * 1000. / 3600.;
		}
		else if(Rand.u()< outdoor_pedestrian_portion/(1- indoor_ue_portion))
		{
			ue_indoor_outdoor_state = PEDESTRIAN;
			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}
		else
		{
			ue_indoor_outdoor_state = IN_CAR;
			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}
	}

	double theta_v = PI/2.;
	double phi_v = 2. * PI*Rand.u();
	
	
	velocity.x = speed * sin(theta_v)*cos(phi_v);
	velocity.y = speed * sin(theta_v)*sin(phi_v);
	velocity.z = speed * cos(theta_v);

	
	if (test_environment >= 3)
	{
		if (ue_indoor_outdoor_state == INDOOR)
		{
			Set_2D_distance_indoor();

			if (Rand.u() < high_loss_type_portion)
			{
				O2I_penetration_loss_model = High_loss_model;
			}
			else
			{
				O2I_penetration_loss_model = Low_loss_model;
			}
		}
		else if (ue_indoor_outdoor_state == IN_CAR)
		{
			Set_car_penet_loss();
		}
	}
		

	if ((ue_indoor_outdoor_state==INDOOR) && (test_environment>=3))
	{
		Calculate_O2I_penet_loss(_2D_distance_indoor);
	}
	
}


/*===================================================================
FUNCTION: UE::Set_2D_distance_indoor( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_2D_distance_indoor

NOTES: m.eval2020 page 44,45

===================================================================*/
void UE::Set_2D_distance_indoor()
{	
	if (pathloss_model == UMa_A || pathloss_model == UMi_A)
	{
		if (carrier_freq < 6000000000)
		{
			_2D_distance_indoor = Rand.u() * 25;
		}
		else
		{
			_2D_distance_indoor = MIN(Rand.u() * 25, Rand.u() * 25);
		}

	}
	else if (pathloss_model==RMa_A)
	{
		if (carrier_freq < 6000000000)
		{
			_2D_distance_indoor = Rand.u() * 10;
		}
		else
		{
			_2D_distance_indoor = MIN(Rand.u() * 10, Rand.u() * 10);
		}
	}
	else if (pathloss_model == UMa_B || pathloss_model == UMi_B)
	{
		_2D_distance_indoor = MIN(Rand.u() * 25, Rand.u() * 25);
	}
	else if (pathloss_model == RMa_B)
	{
		_2D_distance_indoor = MIN(Rand.u() * 10, Rand.u() * 10);
	}


}



/*===================================================================
FUNCTION: UE::Set_O2I_penet_loss( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_O2I_penet_loss

NOTES: m.eval2020 page 44,45

===================================================================*/

void UE::Calculate_O2I_penet_loss(double distance_indoor)
{	

	if (ue_indoor_outdoor_state == INDOOR)
	{
		if ((pathloss_model == UMa_A || pathloss_model == UMi_A) && carrier_freq < 6000000000)
		{
			Pathloss_tw = 20 + 0.5*distance_indoor;
		}
		else if (pathloss_model == RMa_A && carrier_freq < 6000000000)
		{
			Pathloss_tw = 10 + 0.5*distance_indoor;
		}
		else
		{
			if (O2I_penetration_loss_model == High_loss_model)
			{

				Pathloss_tw = High_loss_model_PL + 0.5*distance_indoor +High_loss_model_std* Rand.n();
			}
			else
			{
				Pathloss_tw = Low_loss_model_PL + 0.5*distance_indoor +Low_loss_model_std*Rand.n();
	
			}
		}
	}
	else
	{
		Pathloss_tw = 0;
	}

}


/*===================================================================
FUNCTION: UE::Set_car_penet_loss( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_car_penet_loss

NOTES: m.eval2020 page 45

===================================================================*/

void UE::Set_car_penet_loss()
{
	Pathloss_car = car_penet_loss_mean + car_penet_loss_std*Rand.n();
}


/*===================================================================
FUNCTION: UE::Set_h_UT( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_h_UT

NOTES: m.eval2020 page 19 Table 8-X1 UE antenna height

===================================================================*/
void UE::Set_h_UT()
{
	double n_fl;
	double N_fl;


	if (test_environment >= 3 && test_environment < 7)
	{
		if (ue_indoor_outdoor_state == INDOOR)
		{
			N_fl = floor(5.*Rand.u() + 4);
			n_fl = floor(N_fl*Rand.u() + 1);
			h_UT = 3. * (n_fl - 1) + 1.5;
			floor_idx = int(n_fl) - 1;

		}
		else
		{
			h_UT = 1.5;
			floor_idx = 0;
		}
	}
	else
	{
		floor_idx = 0;
		h_UT = 1.5;
	}
}


/*===================================================================
FUNCTION: Drop_ue_in_cell()

AUTHOR: G.J

DESCRIPTION: Drop ues

NOTES: UE  sector Ȥ cell ȿ ߸.

OUTPUT : location_x, location_y

===================================================================*/
void UE::Drop_ue_in_site()
{

	if (test_environment == Dense_Urban_eMBB_C_Micro)
	{
		do
		{
			location_in_site.x = micro_cell_radius * Rand.u();
			location_in_site.y = micro_cell_radius * Rand.u();

		} while ((sqrt(location_in_site.x*location_in_site.x + location_in_site.y*location_in_site.y) < secter_to_ue_min_distance) || (sqrt(location_in_site.x*location_in_site.x + location_in_site.y*location_in_site.y) > micro_cell_radius));

		double x = Rand.u();
		if (x < 0.25)
		{
			location_in_site.y = -location_in_site.y;
		}
		else if (x < 0.5)
		{
			location_in_site.x = -location_in_site.x;
			location_in_site.y = -location_in_site.y;
		}
		else if (x < 0.75)
		{
			location_in_site.x = -location_in_site.x;
		}
	}
	else
	{

		double x;
		double y;
		
		if ((_ue_idx/num_user_per_sector) % 3 == 0)
		{
			
			do
			{
				x = inter_site_distance / sqrt(3.)* Rand.u();
				y = inter_site_distance * (1. / 2. + 1. / 4.)* Rand.u() - 1. / 4.*inter_site_distance;
			} while ((y > -sqrt(3.)*(x - 1. / sqrt(3.)*inter_site_distance)) || (y < sqrt(3.)*(x - 1. / sqrt(3.)*inter_site_distance)) || (y < -1. / sqrt(3.)*(x)) || (sqrt(x*x + y*y) < secter_to_ue_min_distance));
		}
		else if ((_ue_idx / num_user_per_sector) % 3 == 1)
		{
			
			do
			{
				x = - inter_site_distance / sqrt(3.)* Rand.u();
				y = inter_site_distance * (1. / 2. + 1. / 4.)* Rand.u() - 1. / 4.*inter_site_distance;
			} while ((y > sqrt(3.)*(x + 1. / sqrt(3.)*inter_site_distance)) || (y < -sqrt(3.)*(x + 1. / sqrt(3.)*inter_site_distance)) || (y < 1. / sqrt(3.)*(x)) || (sqrt(x*x + y * y) < secter_to_ue_min_distance));
		}
		else
		{
			
			do
			{
				x = inter_site_distance * (sqrt(3.) / 2)* Rand.u() - inter_site_distance* (sqrt(3.) / 4);
				y = -inter_site_distance / 2.* Rand.u();

			} while ( (y < -sqrt(3.)*(x + 1. / sqrt(3.)*inter_site_distance)) || (y < sqrt(3.)*(x - 1. / sqrt(3.)*inter_site_distance)) || (y > 1. / sqrt(3.)*(x)) || (y > -1. / sqrt(3.)*(x)) || (sqrt(x*x + y * y) < secter_to_ue_min_distance));
		}
		
		location_in_site.x = x;
		location_in_site.y = y;
		//cout << pos_in_site.y << endl;
		
		/*
		bool a = 0;

		if (a == 0)
		{
			do
			{
				pos_in_cell.x = inter_site_distance / 2.* Rand.u();
				pos_in_cell.y = inter_site_distance / sqrt(3.)* Rand.u();
			} while (pos_in_cell.y > (-pos_in_cell.x + inter_site_distance) / sqrt(3.) || sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance);
		}
		else
		{
			do
			{
				pos_in_cell.y = inter_site_distance / 2.* Rand.u();
				pos_in_cell.x = inter_site_distance / sqrt(3.)* Rand.u();
			} while (pos_in_cell.x > (-pos_in_cell.y + inter_site_distance) / sqrt(3.)	|| sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance);

		}

		double x = Rand.u();

		if (x < 0.25)
		{
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.5)
		{
			pos_in_cell.x = -pos_in_cell.x;
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.75)
		{
			pos_in_cell.x = -pos_in_cell.x;
		}
		*/
	}
}

/*===================================================================
FUNCTION: Determine_LOS()

AUTHOR: G.J

DESCRIPTION: Determine_LOS

NOTES: m.2020 eval page 46

OUTPUT : ue[ue_idx].los[sec_idx]

===================================================================*/
void UE::Determine_LOS(int bs_idx)
{
	double d_2D =  channel[bs_idx][_ue_idx]._2D_distance;
	double d_in=0;

	if ((test_environment>=3) && (ue_indoor_outdoor_state == INDOOR))
	{
		if (d_2D < _2D_distance_indoor)
		{			
			d_in = d_2D;
		}
		else
		{
			d_in = _2D_distance_indoor;
		}
	}
	else
	{
		d_in = 0;
	}
	double d_2D_out = d_2D - d_in;
	double P_LOS;



	if (test_environment <3)
	{
		if (d_2D < 5)
		{
			P_LOS = 1.0;
		}
		else if(d_2D < 49)
		{
			P_LOS = exp(-(d_2D - 5.) / 70.8);
		}
		else
		{
			P_LOS = exp(-(d_2D - 49.) / 211.7)*0.54;
		}
	}
	else if (test_environment<6)
	{
		if (d_2D <= 18.0)
		{
			P_LOS = 1.0;
		}
		else
		{
			double C;
				
			if (h_UT <= 13.0)
			{
				C = 0;
			}
			else
			{
				C = pow((h_UT - 13.0) / 10.0, 1.5);
			}

			if (ue_indoor_outdoor_state== INDOOR)
			{
				P_LOS = (18.0 / d_2D_out + exp(-d_2D_out / 63)*(1 - 18 / d_2D_out)) * (1 + C*5.0 / 4.0*pow(d_2D_out / 100.0, 3)*exp(-d_2D_out / 150.0));
			}
			else
			{
				P_LOS = (18.0 / d_2D + exp(-d_2D / 63)*(1 - 18 / d_2D)) * (1 + C*5.0 / 4.0*pow(d_2D / 100.0, 3)*exp(-d_2D / 150.0));
			}
		}
	}
	else if (test_environment == 6)
	{
		if (d_2D <= 18.0)
		{
			P_LOS = 1.0;
		}
		else
		{
			if (ue_indoor_outdoor_state == INDOOR)
			{
				P_LOS = (18.0 / d_2D_out + exp(-d_2D_out / 36)*(1 - 18 / d_2D_out));
			}
			else
			{
				P_LOS = (18.0 / d_2D + exp(-d_2D / 63)*(1 - 18 / d_2D));
			}
		}
	}
	else
	{
		if (d_2D <= 10.)
		{
			P_LOS = 1.0;
		}
		else
		{
			if (ue_indoor_outdoor_state == INDOOR)
			{
				P_LOS = exp(-(d_2D_out - 10)/1000);
			}
			else
			{
				P_LOS = exp(-(d_2D - 10) / 1000);
			}
		}

	}
	

	if ( Rand.u() < P_LOS)
	{
		channel[bs_idx][_ue_idx].propagation_condition = LOS;
		channel[bs_idx][_ue_idx].fast_fading_condition = LOS;
	}
	else
	{
		channel[bs_idx][_ue_idx].propagation_condition = NLOS;
		channel[bs_idx][_ue_idx].fast_fading_condition = NLOS;
		
	}
	
	
	if ((test_environment>=3)&&(ue_indoor_outdoor_state == INDOOR))
	{
		channel[bs_idx][_ue_idx].fast_fading_condition = OUT2IN;
	}

	
	 //ue_indoor_outdoor_state = IN_CAR;

	//LOS/NLOS
	//channel[bs_idx][_ue_idx].propagation_condition = LOS;
	//channel[bs_idx][_ue_idx].fast_fading_condition = LOS;
	

}


/*===================================================================
FUNCTION: void UE::Set_breakpoint_distance(int sec_idx)

AUTHOR: G.J

DESCRIPTION: Set_breakpoint_distance

NOTES: 

OUTPUT : ue[ue_idx].break_point_distance[sec_idx]

===================================================================*/
void UE::Set_breakpoint_distance(int bs_idx)
{
	double d_2D = channel[bs_idx][_ue_idx]._2D_distance;;

	if (test_environment<7)// not rural
	{
		double C;
		double g;
		double h_E;


		if (d_2D <= 18.)
		{
			g = 0.;
		}
		else
		{
			g = 5. / 4. * pow(d_2D / 100., 3.)*exp(-d_2D / 150.);
		}

		if (h_UT < 13.)
		{
			C = 0;
		}
		else
		{
			C = pow((h_UT - 13.) / 10., 1.5)*g;
		}


		if (Rand.u() < 1 / (1 + C))
		{
			h_E = 1.;
		}
		else
		{
			h_E = floor((h_UT - 1.5 - 9.) / 3. * Rand.u()) * 3. + 12.;
		}

		channel[bs_idx][_ue_idx].break_point_distance = 4. * (h_BS - h_E) * (h_UT - h_E) * carrier_freq / SPEED_LIGHT;

	}
	else
	{		
		channel[bs_idx][_ue_idx].break_point_distance = 2. * PI * h_BS * h_UT * carrier_freq / SPEED_LIGHT;
	}
}


/*===================================================================
FUNCTION: void UE::Compute_PathLoss(int sec_idx)

AUTHOR: G.J

DESCRIPTION: Compute_PathLoss

NOTES:

OUTPUT : ue[ue_idx].pathloss[bs_idx]

===================================================================*/
void UE::Compute_PathLoss(int bs_idx)
{
	double d_2D = channel[bs_idx][_ue_idx]._2D_distance;
	double d_in = 0;


	double d_3D = pow(pow(d_2D, 2.) + pow(h_BS - h_UT, 2.), 0.5);
	double d_BP = channel[bs_idx][_ue_idx].break_point_distance;
	double fc = carrier_freq / 1000000000.;
	
	channel[bs_idx][_ue_idx]._3D_distance = d_3D;
	
	

	if (channel[bs_idx][_ue_idx].propagation_condition == LOS)
	{
		if (pathloss_model == InH_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.8 + 16.9*log10(d_3D) + 20 * log10(fc);

					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_A " << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 17.3*log10(d_3D) + 20 * log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_A " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc, LOS Compute_PathLoss pathloss_model InH_A" << endl;
				getchar();
			}
		}
		else if (pathloss_model == InH_B)
		{
			if (fc >= 0.5 &&fc <= 100)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 17.3*log10(d_3D) + 20 * log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc, LOS Compute_PathLoss pathloss_model InH_B" << endl;
				getchar();
			}

		}
		else if (pathloss_model == UMa_A)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					
					cout << "check d_2D, LOS Compute_PathLoss UMa_A " << endl;
					getchar();
				}

			
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMa_A" << endl;
				getchar();
			}
			
		}
		else if (pathloss_model == UMa_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMa_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMa_B" << endl;
				getchar();
			}
		}
		else if (pathloss_model == UMi_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 22.*log10(d_3D) + 28.0 + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_A " << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_A " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMi_A" << endl;
				getchar();
			}

		}
		else if (pathloss_model == UMi_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMi_B" << endl;
				getchar();
			}
		}
		else if (pathloss_model == RMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D;
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					channel[bs_idx][_ue_idx].pathloss_dB = PL1 + 40 * log10(d_3D / d_BP);
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss RMa_A " << endl;
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model RMa_A" << endl;
				getchar();
			}

		}
		else if (pathloss_model == RMa_B)
		{
			if (fc >= 0.5 && fc <= 30)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D;
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					channel[bs_idx][_ue_idx].pathloss_dB = PL1 + 40 * log10(d_3D / d_BP) ;
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss RMa_B " << endl;
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model RMa_B" << endl;
				getchar();
			}
		}

		
	}
	else
	{
		double PL_LOS;
		double PL_NLOS;

		if (pathloss_model == InH_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 43.3*log10(d_3D) + 11.5 + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
					cout << "bs_idx : "<< bs_idx << endl;
					cout << "ue_idx : "<<_ue_idx << endl;
					cout << channel[bs_idx][_ue_idx]._2D_distance << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (d_2D <= 150)
				{
					PL_LOS = 32.4 + 17.3*log10(d_3D) + 20.0*log10(fc);
					PL_NLOS = 17.3 + 38.3*log10(d_3D) + 24.9*log10(fc);
					channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
					if(PL_LOS>PL_NLOS)
					{
						channel[bs_idx][_ue_idx].sigma_SF = 3;
					}
					else
					{
						channel[bs_idx][_ue_idx].sigma_SF = 8.03;
					}
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss InH_A" << endl;
			}
		}
		else if (pathloss_model == InH_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (d_2D <= 150.)
				{
					PL_LOS = 32.4 + 17.3*log10(d_3D) + 20.0*log10(fc);
					PL_NLOS = 17.3 + 38.3*log10(d_3D) + 24.9*log10(fc);
					channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
					if (PL_LOS>PL_NLOS)
					{
						channel[bs_idx][_ue_idx].sigma_SF = 3;
					}
					else
					{
						channel[bs_idx][_ue_idx].sigma_SF = 8.03;
					}
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss InH_A" << endl;
			}
		}
		else if (pathloss_model == UMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMa_A " << endl;
				}

				if (10. <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3)
						+ 20.*log10(fc) - (3.2*pow(log10(17.625), 2) - 4.97) - 0.6*(h_UT - 1.5);
				}
				else
				{

					cout << "d_2D : " << d_2D << endl;
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				

				
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);

				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}

			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 13.54 + 39.08*log10(d_3D) + 20 * log10(fc) - 0.6 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMa_A" << endl;
			}

		}
		else if (pathloss_model == UMa_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_B " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 13.54 + 39.08*log10(d_3D) + 20 * log10(fc) - 0.6 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_B " << endl;
				}
				
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
			/*	cout << h_BS << endl;
				cout << h_UT << endl;
				cout << "d_BP :  "<< d_BP << endl;
				cout << "PL_LOS : " << PL_LOS << endl;
				cout << "PL_NLOS: " << PL_NLOS << endl;
				cout << -channel[bs_idx][_ue_idx].pathloss_dB << endl;
				getchar();*/

				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMa_B" << endl;
			}



		}
		else if (pathloss_model == UMi_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc) ;
				}
				else if (d_BP < d_2D	&& d_2D < 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.) // !!!!!! Table 3.3 10m<d_2d<2000m ,
				{
					PL_NLOS = 36.7*log10(d_3D) + 22.7 + 26 * log10(fc) - 0.3*(h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB= MAX(PL_LOS, PL_NLOS);

				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}

			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP < d_2D	&& d_2D < 5000.)
				{
					PL_LOS = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}

				if (secter_to_ue_min_distance < d_2D && d_2D < 5000.)   ///!! table3-3 shadowing check in 38.901 7.82
				{
					PL_NLOS = 35.3*log10(d_3D) + 22.4 + 21.3 * log10(fc) - 0.3 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 7.82;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMi_A" << endl;
			}
		}
		else if (pathloss_model == UMi_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_B " << endl;
					getchar();
				}

				if (10. < d_2D && d_2D < 5000.)  ///!! table3-3 shadowing check
				{
					PL_NLOS = 35.3*log10(d_3D) + 22.4 + 21.3* log10(fc) - 0.3 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 7.82;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMi_A" << endl;
			}
		}
		else if (pathloss_model == RMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= 21000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3) + 20 * log10(fc) - (3.2*pow(log10(11.75*h_UT), 2) - 4.97);
					channel[bs_idx][_ue_idx].sigma_SF = 8;
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss RMa_A" << endl;
			}
		}
		else if (pathloss_model == RMa_B)
		{
			if (fc >= 0.5 && fc <= 30)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D ;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					PL_LOS = PL1 + 40 * log10(d_3D / d_BP);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_B " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 21000.)
				{
					PL_NLOS = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3) + 20 * log10(fc) - (3.2*pow(log10(11.75*h_UT), 2) - 4.97);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_B " << endl;
				}

				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				
				if (PL_LOS>PL_NLOS)
				{
					if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
					{
						channel[bs_idx][_ue_idx].sigma_SF = 4;
					}
					else
					{
						channel[bs_idx][_ue_idx].sigma_SF = 6;
					}
					
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 8;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss RMa_A" << endl;
			}
		}


	}

	if (channel[bs_idx][_ue_idx].fast_fading_condition == OUT2IN)
	{
		if (pathloss_model == RMa_A || pathloss_model == RMa_B)
		{
			channel[bs_idx][_ue_idx].sigma_SF = 8;
		}
		else
		{
			channel[bs_idx][_ue_idx].sigma_SF = 7;
		}
		
	}

	channel[bs_idx][_ue_idx].pathloss_dB += Pathloss_tw+Pathloss_car;

	channel[bs_idx][_ue_idx].pathloss = dB2linear(-channel[bs_idx][_ue_idx].pathloss_dB);

}


void UE::CSI_calculation()
{
	double mutual_inf[4] = { 0,0,0,0};
	

	int PMI_set[4][52]; //[layer][num_rb];
	
	if (t == 0)
	{
		for (int num_layer = 0; num_layer<num_ue_antenna_port; num_layer++) // 
		{
		
			for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
			{
				PMI_set[num_layer][rb_idx] = Find_PMI( mutual_inf, SINR_dB, num_layer, rb_idx);
			}	

			//cout <<"mutual information["<< num_layer  <<"] : "<< mutual_inf[num_layer]/num_rb << endl;
		}

		RI = 0;
		for (int num_layer = 0; num_layer < num_ue_antenna_port; num_layer++)
		{
			if (mutual_inf[RI]< mutual_inf[num_layer])
			{
				RI = num_layer;
			}
		}

		for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
		{
			for (int idx = 0; idx < csi_history_length ; idx++)
			{

				estimated_eff_channel[rb_idx][idx] = MatrixXcd(bs_antenna.num_txru, RI + 1);
			}
		}
	}
	else
	{
		
		for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
		{
			PMI_set[RI][rb_idx] = Find_PMI(mutual_inf, SINR_dB, RI, rb_idx);
		}
		/*if (_ue_idx == 61)
		{
			cout << "_ue_idx : " << _ue_idx << endl;

			for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
			{
				cout << "layer_idx : " << layer_idx << endl;

				for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
				{
					cout << "SINR[" << rb_idx << "]: " << SINR_dB[RI][layer_idx][rb_idx] << endl;
				}
			}
			getchar();
		}
		*/
	}
	
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		PMI[rb_idx][t%csi_history_length] = PMI_set[RI][rb_idx];
	}

	Find_R_for_pf(SINR_dB);

	//for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	//{
	//	CQI[rb_idx][t%csi_history_length] = Map_CQI(ESINR_dB);
	//}


}

void UE::Calculate_Interference()
{
	int ue_selected;
	int RI;
	int PMI;
	int bs_idx = sector_in_control/num_TRxP_per_site;
	
	MatrixXcd H;
	MatrixXcd W;
	MatrixXcd G;

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		R_inter[rb_idx].setZero();
	}

	double interference = 0;

	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		for (int sec_idx = 0; sec_idx < num_sector; sec_idx++)
		{
			if (sec_idx != sector_in_control)
			{
				ue_selected = sector[sec_idx].pschedulewrite[rb_idx].ue_selected;
				RI = sector[sec_idx].pschedulewrite[rb_idx].RI;
				PMI = sector[sec_idx].pschedulewrite[rb_idx].PMI;

				H = channel[bs_idx][ue_selected].H[sector_in_control%num_TRxP_per_site][rb_idx];
				if (num_ue_antenna_port == 4)
				{
					if (RI == 0) { W = W_sigle_layer[PMI]; }
					else if (RI == 1) { W = W_two_layer[PMI]; }
					else if (RI == 2) { W = W_three_layer[PMI]; }
					else if (RI == 3) { W = W_four_layer[PMI]; }
					else
					{
						cout << "sector[" << sec_idx << "].pschedulewrite[" << rb_idx << "].RI : " << sector[sec_idx].pschedulewrite[rb_idx].RI << endl;
						cout << endl;
						cout << "check RI" << endl;
						getchar();
					}

				}
				else if (num_ue_antenna_port == 2)
				{
					if (RI == 0) { W = W_sigle_layer_two_port[PMI]; }
					else if (RI == 1) { W = W_two_layer_two_port[PMI]; }
				}

				G = H * W;
				
				R_inter[rb_idx] += ue[ue_selected].power_per_RB/(double(RI)+1) * channel[bs_idx][ue_selected].SF * channel[bs_idx][ue_selected].pathloss  *(G * G.adjoint());
			
			}
			
		}
		
	}

}

int UE::Find_PMI(double mutual_inf[], double SINR[][4][52], int num_layer, int rb_idx)
{

	double num_PMI[4] = { 28,21,7,5 };
	
	double temp_mu_Inf;
	double max_mu_inf;
	int bs_idx = sector_in_control / num_TRxP_per_site;
	int sector_idx = sector_in_control % num_TRxP_per_site;
	int max_PMI;
	double P0 = Update_Power_Control(num_rb);
	double Received_power = channel[bs_idx][_ue_idx].SF * channel[bs_idx][_ue_idx].pathloss*P0;
	MatrixXcd R_thermal(bs_antenna.num_txru, bs_antenna.num_txru);
	MatrixXcd R_intra(bs_antenna.num_txru, bs_antenna.num_txru);;
	
	double signal = 0;
	double temp_SINR[4];
	double interference = 0;
	
	double max_signal = 0;
	double max_interference = 0;

	
	R_thermal.setIdentity();

	R_thermal = R_thermal * dBm2linear(thermal_nosie_level + 10* log10(subcarrier_spacing * 12) + bs_noise_figure);
	
	

	for (int PMI_idx=0; PMI_idx<  num_PMI[num_layer]; PMI_idx++)
	{		
		MatrixXcd G; 
		MatrixXcd W_mmse;
		temp_mu_Inf = 0;
		if (num_ue_antenna_port == 4)
		{
			if (num_layer == 0) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_sigle_layer[PMI_idx]; }
			else if (num_layer == 1) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_two_layer[PMI_idx]; }
			else if (num_layer == 2) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_three_layer[PMI_idx]; }
			else if (num_layer == 3) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_four_layer[PMI_idx]; }
			else
			{
				cout << "check int UE::Find_PMI(double mutual_inf[], double SINR[][4][52], int num_layer, int rb_idx)" << endl;
				getchar();
			}
		}
		else if (num_ue_antenna_port == 2)
		{
			if (num_layer == 0) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_sigle_layer_two_port[PMI_idx]; }
			else if (num_layer == 1) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_two_layer_two_port[PMI_idx]; }
		}
		
		W_mmse = Received_power / (num_layer + 1)*G.adjoint()*(Received_power / (num_layer + 1)*(G*G.adjoint()) +R_thermal).inverse();//+ R_inter[rb_idx]


		for (int layer_idx = 0; layer_idx < num_layer + 1; layer_idx++)
		{
			R_intra.setZero();
			for (int intra_idx = 0; intra_idx < num_layer + 1; intra_idx++)
			{
				if (layer_idx != intra_idx)
				{		
					R_intra += Received_power / (num_layer + 1) * ( G.col(intra_idx) * G.col(intra_idx).adjoint() );
				}
			}

			interference = (W_mmse.row(layer_idx)*(R_intra 	+ R_thermal)* W_mmse.row(layer_idx).adjoint())(0).real();//+ R_inter[rb_idx] 

			signal =( Received_power / (num_layer + 1))  * (W_mmse.row(layer_idx)*G.col(layer_idx)*(W_mmse.row(layer_idx)*G.col(layer_idx)).adjoint())(0).real();

			
			temp_SINR[layer_idx] = signal / interference;

			temp_mu_Inf += log2(1. + temp_SINR[layer_idx]);
		}

		if (PMI_idx == 0)
		{
			for (int layer_idx = 0; layer_idx < num_layer + 1; layer_idx++)
			{
				SINR[num_layer][layer_idx][rb_idx] = temp_SINR[layer_idx];
			}
			max_signal = signal;
			max_interference = interference;
			max_mu_inf =temp_mu_Inf;
			max_PMI = 0;

		}
		else if (max_mu_inf < temp_mu_Inf)
		{
			for (int layer_idx = 0; layer_idx < num_layer + 1; layer_idx++)
			{
				SINR[num_layer][layer_idx][rb_idx] = temp_SINR[layer_idx];
			}

			max_signal = signal;
			max_interference = interference;

			max_mu_inf = temp_mu_Inf;
			max_PMI = PMI_idx;
		}
	
	}

	//if (_ue_idx == 61)
	//{
	//	cout << "power : " << Received_power << endl;
	//	cout << "signal : " << max_signal << endl;
	//	cout << "interference :" << max_interference << endl;
	//}


	mutual_inf[num_layer] += max_mu_inf;

	return max_PMI;
}

void UE::Find_R_for_pf(double SINR[][4][52])
{
	double si_sum = 0;
	double RBIR;
	double ESINR_dB;
//	cout << endl;
	
	int num_symbol = 0;
	for (int rb_idx = 0; rb_idx < rs_size; rb_idx++)
	{
		si_sum = 0;
		for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
		{
				si_sum += Mapping_RBIR_SNR(2, SINR[RI][layer_idx][rb_idx], 0); // int mod_order : 0 QPSK, 1 16QAM , 2 64QAM  double input : SINR or RBIR,  int mode :  case 0 SINR->RBIR, case 1 RBIR->ESINR
		}
		RBIR = si_sum / double(RI + 1);
		ESINR_dB = Mapping_RBIR_SNR(2, RBIR, 1);
		R_for_pf[rb_idx] = log2(1 + dB2linear(ESINR_dB));
	}

}



/*===================================================================
FUNCTION: UE::Update_Power_Control( void )

AUTHOR: JH

DESCRIPTION: return the total throughput over the frequency

NOTES:

===================================================================*/
double UE::Update_Power_Control(int num_of_rbs)
{
	
	if (num_of_rbs == 0)
	{
		return 0.0;
	}
	else
	{
		double P_max = ue_power;



		int M_pusch = num_of_rbs;
		double Ks = 1.25;
		int N_info = info_bits;
		int N_re = num_of_rbs * 12 * 14;
		double MPR = (double)N_info / (double)N_re;
		double delta_TF = 0.;// 10*log10( pow( 2. , MPR*Ks) - 1);
		double f_of_i = 0.;
		
		double Total_PL = channel[sector_in_control/ num_TRxP_per_site][_ue_idx].pathloss_dB;  //  ʿ-channel[sector_in_control][_ue_idx].LS_gain .
		double power;
		double power_per_rb_dBm;

		power = MIN(P_max , Power_0 + 10. * log10(M_pusch) + Power_alpha * Total_PL + delta_TF + f_of_i);
		
		power_per_rb_dBm = power - 10.*log10(M_pusch);

		
		
		return dBm2linear(power_per_rb_dBm); // (num_tx_antenna_element);
		//return 0.0;

	}
}

void UE::Receive_UL()
{
	double instantThroughput = 0;

	if (rb_indices.size() > 0)
	{
		if( ack== ACK){num_transmission++;}
		
		//cout << "Compute_Effective_Snr" << endl;
		esinr =Compute_Effective_Snr(1);

		if (ack == NACK)
		{
			cout <<"num_retransmission : "<< num_retransmission << endl;
			cout << "num_of_rbs[" << _ue_idx << "] : " << num_of_rbs << endl;
			cout << "CQI["<< _ue_idx <<"] : "<<CQI_selected << endl;
			cout << "esinr[" << _ue_idx << "] : " <<linear2dB( esinr) << endl;
			cout << endl;
		}
		//if(_ue_idx==564)
		//{
		//	cout << "ESINR[" << _ue_idx << "] : " << linear2dB(esinr) << endl;
		//	cout << "CQI_selected : " << CQI_selected << endl;
		//}
		
		//spectralEfficiency = Get_SpectralEfficiency(esinr, mcs);
		
		//cout << "FER(esinr, CQI_selected)" << endl;
		if (FER(esinr, CQI_selected)< Rand.u())
		{
			ack = ACK;
			num_retransmission = 0;
			HARQeSINRs.clear();
			//cout << "Determine_TBS" << endl;
			instantThroughput = Determine_TBS(RI+1, CQI_selected, num_of_rbs);
		}
		else
		{
			ack = NACK;
			num_retransmission++;
			if (num_retransmission == 1){	num_error[0]++;	}
			else if (num_retransmission == 2) { num_error[1]++; }
			else if (num_retransmission == 3) { num_error[2]++; }
			else if (num_retransmission == 4) { num_error[3]++; }
			else
			{
				ack = ACK; // fail
			}

			instantThroughput = 0;

		}


	
	}

	if (instantThroughput < 0)
	{
		cout << "check instantThroughput 1 " << endl;
		cout << "instantThroughput :" << instantThroughput << endl;
		cout << "RI :" << RI << endl;
		cout << "CQI_selected :" << CQI_selected << endl;
		cout << "num_of_rbs :" << num_of_rbs << endl;
		getchar();
	}

	uplinkThroghput = uplinkThroghput  + instantThroughput ;

	//cout << uplinkThroghput << endl;
}


void UE::Find_Allocated_Rbs(void)
{
	rb_indices.clear();
	num_of_rbs = 0;
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		if (sector[sector_in_control].pschedulewrite[rb_idx].ue_selected == _ue_idx)
		{
			rb_indices.push_back(rb_idx);
		}
	}
	num_of_rbs = int(rb_indices.size());

	//cout << "num_of_rbs : " << num_of_rbs << endl;
	//getchar();
}


double UE::Compute_Effective_Snr(int mode) // return linear SINR
{
	double P0 = power_per_RB;
	int bs_idx = sector_in_control / num_TRxP_per_site;
	int sector_idx = sector_in_control % num_TRxP_per_site;
	double Received_power = channel[bs_idx][_ue_idx].SF * channel[bs_idx][_ue_idx].pathloss*P0;
	MatrixXcd R_thermal(bs_antenna.num_txru, bs_antenna.num_txru);
	MatrixXcd R_intra(bs_antenna.num_txru, bs_antenna.num_txru);;
	MatrixXcd G;
	MatrixXcd W_mmse;

	double signal = 0;
	double interference = 0;
	double SINR[4][52];
	
	R_thermal.setIdentity();
	R_thermal = R_thermal * dBm2linear(thermal_nosie_level + 10 * log10(subcarrier_spacing * 12) + bs_noise_figure);
	
	//cout << "ue_idx : " << _ue_idx << endl;
	//cout << "num_of_rbs : " << num_of_rbs << endl;
	//cout << "RI : " << RI << endl;
	//
	//for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	//{
	//	cout << estimated_eff_channel[rb_idx][t%csi_history_length] << endl;
	//}

	for (int i = 0; i < num_of_rbs; i++)
	{

		for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
		{
			if (rb_indices[i] == rb_idx)
			{
				int PMI = sector[sector_in_control].pschedulewrite[rb_idx].PMI;
				if (mode == 1)
				{
					if (num_ue_antenna_port == 4)
					{
						if (RI == 0) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_sigle_layer[PMI]; }
						else if (RI == 1) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_two_layer[PMI]; }
						else if (RI == 2) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_three_layer[PMI]; }
						else if (RI == 3) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_four_layer[PMI]; }
					}
					else if (num_ue_antenna_port == 2)
					{
						if (RI == 0) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_sigle_layer_two_port[PMI]; }
						else if (RI == 1) { G = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_two_layer_two_port[PMI]; }
					}
				}
				else
				{
					if (num_ue_antenna_port == 4)
					{
						if (RI == 0) { G = estimated_eff_channel[rb_idx][t%csi_history_length]; }
						else if (RI == 1) { G = estimated_eff_channel[rb_idx][t%csi_history_length]; }
						else if (RI == 2) { G = estimated_eff_channel[rb_idx][t%csi_history_length]; }
						else if (RI == 3) { G = estimated_eff_channel[rb_idx][t%csi_history_length]; }
					}
					else if (num_ue_antenna_port == 2)
					{
						if (RI == 0) { G = estimated_eff_channel[rb_idx][t%csi_history_length]; }
						else if (RI == 1) { G = estimated_eff_channel[rb_idx][t%csi_history_length]; }
					}

				}
				
				
				if (mode == 1) { W_mmse = Received_power / (RI + 1)*G.adjoint()*(Received_power / (RI + 1)*(G*G.adjoint()) + R_thermal).inverse(); }// + R_inter[rb_idx] 
				else  { W_mmse = Received_power / (RI + 1)*G.adjoint()*(Received_power / (RI + 1)*(G*G.adjoint())+ R_thermal).inverse(); }			// + R_inter_history[rb_idx][t%csi_history_length] 
				for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
				{
					R_intra.setZero();
					for (int intra_idx = 0; intra_idx < RI + 1; intra_idx++)
					{
						if (layer_idx != intra_idx)
						{
							R_intra += Received_power / (RI + 1) * (G.col(intra_idx) * G.col(intra_idx).adjoint());
						}
					}

					interference = (W_mmse.row(layer_idx)*(R_intra+ R_thermal)* W_mmse.row(layer_idx).adjoint())(0).real();// + R_inter[rb_idx] 

					signal = (Received_power / (RI + 1))  * (W_mmse.row(layer_idx)*G.col(layer_idx)*(W_mmse.row(layer_idx)*G.col(layer_idx)).adjoint())(0).real();


					
					SINR[layer_idx][i] = signal / interference;

				/*	if (_ue_idx == 564)
					{
						cout << "signal[" << layer_idx << "][" << i << "] : " << signal << endl;
						cout << "interference[" << layer_idx << "][" << i << "] : " << interference << endl;
						cout << "SINR[" << layer_idx << "][" << i << "] : " << SINR[layer_idx][i] << endl;
						cout << endl;
					}*/

					//cout << "SINR["<<layer_idx<<"]["<<i<<"] : "<<  SINR[layer_idx][i] << endl;
				/*	if (_ue_idx == 61)
					{	
						cout << "Received_power :"<< Received_power << endl;
						cout << "signal : " << signal << endl;
						cout << "interference :" << interference << endl;
					}*/
				
				}
			}
		}

	}

/*
	if (_ue_idx == 61)
	{
		cout << "SINR check" << endl;
		cout << "_ue_idx : " << _ue_idx << endl;
		for (int i = 0; i < num_of_rbs; i++)
		{
			for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
			{
				cout << "SINR[" << layer_idx << "][" << i << "]" << SINR[layer_idx][i] << endl;
			}
		}
		getchar();
	}*/

	
	double si_sum = 0;
	double RBIR;
	double ESINR_dB;
	

	if (mode == 1)
	{
		int mod_idx = Get_modulation_Scheme(CQI_selected);
		if (ack == NACK) // chase combining
		{
			/*cout << "if (ack == NACK)  chase combining" << endl;

			cout << "CQI_selected : " << CQI_selected <<endl;
			cout << "num_retx : " << num_retransmission << endl;*/
			for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
			{
				//cout << "layer_idx : " << layer_idx << endl;
				for (int rb_idx = 0; rb_idx < num_of_rbs; rb_idx++)
				{
					/*	cout << "rb_idx : "<<rb_idx << endl;
					cout <<"SINRbefore : "<< SINR[layer_idx][rb_idx] << endl;*/
					SINR[layer_idx][rb_idx] += HARQeSINRs[num_of_rbs*layer_idx + rb_idx];
					//cout << "SINRafter : " << SINR[layer_idx][rb_idx] << endl;	
				}
			}
			//getchar();
		}

		int num_symbol = 0;
		for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
		{
			for (int rb_idx = 0; rb_idx < num_of_rbs; rb_idx++)
			{
				si_sum += Mapping_RBIR_SNR(mod_idx, SINR[layer_idx][rb_idx], 0); // int mod_order : 0 QPSK, 1 16QAM , 2 64QAM  double input : SINR or RBIR,  int mode :  case 0 SINR->RBIR, case 1 RBIR->ESINR
				num_symbol++;
			}
		}
		RBIR = si_sum / double(num_symbol);
		ESINR_dB = Mapping_RBIR_SNR(mod_idx, RBIR, 1);
		//if (ack == NACK) // chase combining
		//{
		//	cout << "ue_idx : " << _ue_idx << endl;
		//	cout << "sector_in_control : " << sector_in_control << endl;
		//	cout << "CQI_selected : " << CQI_selected << endl;
		//	cout << "num_retx : " << num_retransmission << endl;
		//	cout << "ESINR_dB : "<<ESINR_dB << endl;
		//	getchar();
		//}

		HARQeSINRs.clear();

		//cout << " HARQeSINRs.clear();" << endl;
		for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
		{
			for (int rb_idx = 0; rb_idx < num_of_rbs; rb_idx++)
			{
				//cout << "SINR["<<layer_idx<<"]["<< rb_idx <<"] :" <<SINR[layer_idx][rb_idx] << endl;
				HARQeSINRs.push_back(SINR[layer_idx][rb_idx]);
			}
		}

	}
	else
	{
		
		for (int layer_idx = 0; layer_idx < RI + 1; layer_idx++)
		{
			for (int rb_idx = 0; rb_idx < num_of_rbs; rb_idx++)
			{
				si_sum += Mapping_RBIR_SNR(2, SINR[layer_idx][rb_idx], 0); // int mod_order : 0 QPSK, 1 16QAM , 2 64QAM  double input : SINR or RBIR,  int mode :  case 0 SINR->RBIR, case 1 RBIR->ESINR
				
			}
		}
		RBIR = si_sum / double((RI + 1.)*num_of_rbs);

		ESINR_dB = Mapping_RBIR_SNR(2, RBIR, 1);

	}
	
	return dB2linear( ESINR_dB);
}

double UE::FER(double esinr, int mcs)
{
	double Frame_Error_Rate = 1;
	double ESINR_L = esinr; // linear value
	switch (mcs)
	{
	case -1:
		Frame_Error_Rate = 0.0;
		break;

	case 0:
		if (ESINR_L < pow(10.0, -7.737 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(0.9*(73.109 - 460.4 * ESINR_L));
		break;
	case 1:
		if (ESINR_L < pow(10.0, -5.852 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(0.5*(67.572 - 300.63 * ESINR_L));

		break;

	case 2:
		if (ESINR_L < pow(10.0, -3.737 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(0.8*(69.19 - 195.42 * ESINR_L));

		break;

	case 3:
		if (ESINR_L < pow(10.0, -1.718 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(0.8*(70.491 - 113.69 * ESINR_L));

		break;

	case 4:
		if (ESINR_L < pow(10.0, 0.3206 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(1.3*(77.142 - 76.651* ESINR_L));

		break;

	case 5:
		if (ESINR_L < pow(10.0, 2.14 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(1.5*(47.496 - 31.018 * ESINR_L));

		break;

	case 6:
		if (ESINR_L < pow(10.0, 4.096 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(1.3*(46.163 - 16.085 * ESINR_L));

		break;

	case 7:
		if (ESINR_L < pow(10.0, 6.052 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(1.5*(45.936 - 11.402 * ESINR_L));

		break;

	case 8:
		if (ESINR_L < pow(10.0, 8.03 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(1.3*(44.625 - 7.3239 * ESINR_L));

		break;

	case 9:
		if (ESINR_L < pow(10.0, 10.03 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(1.5*(44.766 - 4.8398 * ESINR_L));

		break;

	case 10:
		if (ESINR_L < pow(10.0, 11.82 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(1.5*(41.834 - 2.8525 * ESINR_L));

		break;
	case 11:
		if (ESINR_L < pow(10.0, 13.65 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(2 * (33.542 - 1.5585 * ESINR_L));

		break;
	case 12:
		if (ESINR_L < pow(10.0, 15.69 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(3 * (27.392 - 0.8247 * ESINR_L));

		break;
	case 13:
		if (ESINR_L < pow(10.0, 17.5 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(2 * (28.525 - 0.53504 * ESINR_L));

		break;
	case 14:
		if (ESINR_L < pow(10.0, 18.98 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(5*(8.154 - 0.10315 * ESINR_L));

		break;
	}
	return Frame_Error_Rate;
}



void UE::Save_channel(void)
{
	int bs_idx= sector_in_control / num_TRxP_per_site;;
	int sector_idx= sector_in_control % num_TRxP_per_site;
	
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		if (RI == 0) {estimated_eff_channel[rb_idx][t%csi_history_length] = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_sigle_layer[PMI[rb_idx][t%csi_history_length]]; }
		else if (RI == 1) { estimated_eff_channel[rb_idx][t%csi_history_length] = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_two_layer[PMI[rb_idx][t%csi_history_length]]; }
		else if (RI == 2) { estimated_eff_channel[rb_idx][t%csi_history_length] = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_three_layer[PMI[rb_idx][t%csi_history_length]]; }
		else if (RI == 3) { estimated_eff_channel[rb_idx][t%csi_history_length] = channel[bs_idx][_ue_idx].H[sector_idx][rb_idx] * W_four_layer[PMI[rb_idx][t%csi_history_length]]; }
	}
	for (int rb_idx = 0; rb_idx < num_rb; rb_idx++)
	{
		R_inter_history[t%csi_history_length][rb_idx] = R_inter[rb_idx];
	}
}

void UE::determine_CQI(void)
{
	num_of_rbs = int( rb_indices.size());
	//cout << "ue[" << _ue_idx << "] : " << num_of_rbs << endl;

	if (num_of_rbs > 0 && ack == ACK)
	{
		power_per_RB = Update_Power_Control(num_of_rbs);

		double ESINR;
		ESINR = Compute_Effective_Snr(0);
		//cout <<"ESINR : "<< ESINR << endl;

		CQI_selected = Map_CQI(ESINR);

		//if (_ue_idx == 564)
		//{
		//	cout << "ESINR[" << _ue_idx << "] : " << linear2dB(ESINR) << endl;
		//	cout << "CQI_selected : " << CQI_selected << endl;
		//	cout << "num_rb : " << num_of_rbs << endl;
		//}
		
		//cout << "CQI_selected : " << CQI_selected << endl;
	}
}